#
# Common image characteristics (between MMC and USB)
#
# Turn on splash for customers, by customizing psplash_%.bb which
# is currently customized by meta-raspberrypi
IMAGE_FEATURES += "splash"

# Password hash generated with the following command:
# printf "%q" $(mkpasswd -m sha256crypt INSERT_PASSWORD_HERE)
PASSWD_HASH = "\$5\$z6dHiIfuy6pee1jd\$sUBLGXY7toh5CMg6WQQAa2kMreC8tk6MFwtzV3IVNo8"

# Add kas user and setup groups/passwords
inherit extrausers
EXTRA_USERS_PARAMS = "\
    usermod -p '*' root; \
    useradd -m -p '${PASSWD_HASH}' -G sudo kas; \
    useradd -m -p '${PASSWD_HASH}' -G sudo admin; \
    "

inherit core-image
inherit image-buildinfo

inherit swupdate-image
IMAGE_FSTYPES:append = " ext4.gz"

KAS_OS_VERSION_MAJOR="1"
KAS_OS_VERSION_MINOR="0"
KAS_OS_VERSION_MICRO="0"

# Note: The git commit count and KAS version fields are updated in the anonymous python function.
GIT_COMMIT_COUNT     = "99999" 
KAS_OS_VERSION      ?= ""

IMAGE_NAME_PREFIX    = "${MACHINE_NAME_UPPER}-Cx-08"
IMAGE_VERSION_SUFFIX = "${@oe.utils.conditional("INSTALL_KAS_RUNTIME", "1", \
                             "OS-${KAS_OS_VERSION}-RT-${KAS_RUNTIME_VERSION}", \ 
                             "OS-${KAS_OS_VERSION}", \
                             d)}"
IMAGE_NAME           = "${IMAGE_NAME_PREFIX}-${IMAGE_VERSION_SUFFIX}"
IMAGE_NAME_SUFFIX    = ""

# In BOOTIMAGE_VERSION, "2022.01" is the uboot base version, prior to KAS patches.
# The "08" is the latest patch number in recipes-bsp\u-boot\u-boot_%.bbappend.
UBOOT_VERSION      = "2022.01"
KAS_UBOOT_REVISION = "09"
BOOTIMAGE_VERSION  = "${UBOOT_VERSION}.${KAS_UBOOT_REVISION}"
BOOTIMAGE_SWU_FILE = "${IMAGE_NAME_PREFIX}-bootimage-${BOOTIMAGE_VERSION}.swu"

add_rootfs_version () {
    # Create the KAS login banner file.
 
    # NOTE: for 'git' to succeed, the meta-kas directory must be designated as
    # "safe" for a non-owner of the repo. This is done using git config. Example:
    #     git config --global --add safe.directory /timesys/yocto-kas/layers/meta-kas
    # There may be a fix in Yocto for this in the future. The issue is discussed here:
    # https://lists.openembedded.org/g/openembedded-core/message/165184.
 
    # TODO: The code below to get the KAS_OS_VERSION is redundant. This value is already updated
    # in the python anonymous function (that's executed after parsing the recipes) in this file.
    # Access this variable from the data store instead of recomputing it.
    #           kas_version=${@d.getVar('KAS_OS_VERSION')}
    # Refer to https://stackoverflow.com/questions/48213497/how-do-i-update-a-datastore-variable-from-inside-a-bash-variable
    # 
    GIT_SRC_DIR=${TOPDIR}/../layers/meta-kas
    GIT_COMMIT_COUNT=`git -C ${GIT_SRC_DIR} rev-list HEAD | wc -l`

    KAS_OS_VERSION_MAJOR=1
    KAS_OS_VERSION_MINOR=0
    KAS_OS_VERSION_MICRO=0
    KAS_OS_VERSION="${KAS_OS_VERSION_MAJOR}.${KAS_OS_VERSION_MINOR}.${KAS_OS_VERSION_MICRO}.${GIT_COMMIT_COUNT}"

    echo "OS Version: ${KAS_OS_VERSION}" > ${IMAGE_ROOTFS}/etc/kas_login_banner
    
    printf "${DISTRO_NAME} ${DISTRO_VERSION} (${DISTRO_CODENAME}) \\\n \\\l\n" > ${IMAGE_ROOTFS}/etc/issue
    printf "${DISTRO_NAME} ${DISTRO_VERSION} (${DISTRO_CODENAME}) %%h\n" > ${IMAGE_ROOTFS}/etc/issue.net
    printf "OS Version: ${KAS_OS_VERSION}\n\n" >> ${IMAGE_ROOTFS}/etc/issue
    printf "OS Version: ${KAS_OS_VERSION}\n\n" >> ${IMAGE_ROOTFS}/etc/issue.net

    # Create /etc/kas_version_info.xml
 
    GIT_COMMIT_HASH=`git -C ${GIT_SRC_DIR} rev-parse HEAD`
    GIT_MODIFIED_BY=""
    git -C ${GIT_SRC_DIR} diff --quiet || GIT_MODIFIED_BY=`git -C ${GIT_SRC_DIR} config user.name`

    echo "<?xml version="1.0" encoding="UTF-8" ?>"   > ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "<attributeMap>"                           >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "    <OSVersion "                          >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "        commit_hash=\"$GIT_COMMIT_HASH\"" >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "        modified_by=\"$GIT_MODIFIED_BY\"" >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "        version=\"$KAS_OS_VERSION\""      >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "    />"                                   >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
    echo "</attributeMap>"                          >> ${IMAGE_ROOTFS}/etc/kas_version_info.xml
}

# add the rootfs version to the welcome banner
ROOTFS_POSTPROCESS_COMMAND += "add_rootfs_version;"

python add_boot_version_file() {
    import json

    KAS_UBOOT_REVISION = d.getVar('KAS_UBOOT_REVISION')
    UBOOT_VERSION = d.getVar('UBOOT_VERSION')
    BOOTIMAGE_VERSION = d.getVar('BOOTIMAGE_VERSION')
    DEPLOY_DIR_IMAGE = d.getVar('DEPLOY_DIR_IMAGE')
    BOOTFILES_DIR_NAME = d.getVar('BOOTFILES_DIR_NAME')

    version_dict = {
        "kas_uboot_revision": KAS_UBOOT_REVISION,
        "uboot_version": UBOOT_VERSION,
        "full_uboot_version": BOOTIMAGE_VERSION,
    }

    version_file_path = f'{DEPLOY_DIR_IMAGE}/{BOOTFILES_DIR_NAME}/version.json'

    with open(version_file_path, 'w') as version_file:
        json.dump(
            version_dict, 
            version_file,
            indent = '\t',
            sort_keys = True,
        )
}


# Include kas-runtime in distro.
# specifics of Jenkins build are in kas-runtime.conf
require conf/distro/kas-runtime.conf

# add kas-runtime to rootfs
add_rootfs_kas_runtime () {
   KAS_HOME="${IMAGE_ROOTFS}/home/kas"
   install -d ${KAS_HOME}/kas
   # install -t ${KAS_HOME} ${D}${sysconfdir}/skel/.bashrc ${D}${sysconfdir}/skel/.profile
   tar -C ${KAS_HOME}/kas -xf ${DEPLOY_DIR_TAR}/${KAS_RUNTIME_ARTIFACT_TGZ}
   chmod 4666 ${KAS_HOME}/kas/bin/sinope
}

# has to be called after set_user_group from extrausers.bbclass, so can't
# include it in ROOTFS_POSTPROCESS_COMMAND
do_rootfs[postfuncs] += '${@oe.utils.conditional("INSTALL_KAS_RUNTIME", "1", "add_rootfs_kas_runtime", "", d)}'

# If you want to use u-dev rules instead of the current systemd-based USB automount,
# 1. add "usb-automount" to the list of packages to add to image.
# 2. remove the x-systemd entries from fstab.

IMAGE_INSTALL:append = " \
    bash \
    cifs-utils \
    crashdump-init \
    dool \
    dosfstools \
    e2fsprogs-mke2fs \
    e2fsprogs-resize2fs \
    file \
    gdb \
    gdbserver \
    htop \
    iproute2 \
    ldd \
    libcurl \
    libgpiod \
    libgpiod-dev \
    libgpiod-tools \
    lmsensors \
    lshw \
    lsof \
    lvm2 \
    mbedtls \
    nano \
    network-init \
    ntfsprogs \
    ntfs-3g \
    ntfs-3g-ntfsprogs \
    openssh-sftp \
    openssh-sftp-server \
    parted \
    packagegroup-core-ssh-openssh \
    pcmm-startup \
    pysoem \
    python3 \
    python3-periphery \
    python3-smbus2 \
    python3-pip \
    rtc-battery-check \
    smbclient \
    sudo \
    swupdate \
    swupdate-client \
    tcpdump \
    tzdata \
    u-boot-fw-utils \
    userland \
    watchdog \
    xenomai-evl \
    xmlrpc-c \
    "

# Add following lines if you need to include protobuf and/or grpc in the image:
#   grpc \
#   protobuf \
#   python3-grpcio \
#   python3-grpcio-tools \
#
IMAGE_INSTALL:append = " \
    zeromq \
    cppzmq-dev \
    python3-pyzmq \
    "

# Note: Including kas-runime recipe in the image is optional
IMAGE_INSTALL:append = "${@oe.utils.conditional("INSTALL_KAS_RUNTIME", "1", "kas-runtime", "", d)}"


image_install_rootfs_manifest_file () {
    # Copy manifest file to the image under /etc folder
    if [ -f ${IMAGE_MANIFEST} ]; then
        cp ${IMAGE_MANIFEST} ${IMAGE_ROOTFS}/etc/
    fi
}

image_rootfs_create_default_dir () {
    if [ -d ${IMAGE_ROOTFS}/default ]; then
        rm -rf ${IMAGE_ROOTFS}/default
    fi

    # Create /default directory in the filesystem. 
    # This directory is used to store a copy of the OS configuration files that are needed for factory reset.
    mkdir -p ${IMAGE_ROOTFS}/default

    # Backup /etc directory
    tar -czf ${IMAGE_ROOTFS}/default/etc.tgz --directory ${IMAGE_ROOTFS}/etc .
}

IMAGE_PREPROCESS_COMMAND += " image_install_rootfs_manifest_file; "

do_populate_license_readme() {
    # Add README to the license folder in the rootfs:  /usr/share/common-licenses
    copy_lic_dirs="${@d.getVar("COPY_LIC_DIRS")}"
    rootfs_license_dir="${@d.getVar("ROOTFS_LICENSE_DIR")}"
    bbnote 1 "do_populate_license_readme: copy_lic_dirs: ${copy_lic_dirs}, rootfs_license_dir: ${rootfs_license_dir}"
    if [ "${@d.getVar('COPY_LIC_DIRS')}" = "1" ]; then
        install -m 044 ${WORKDIR}/README ${rootfs_license_dir}
    fi
}
# addtask do_populate_license_readme after do_populate_lic_deploy
IMAGE_PREPROCESS_COMMAND += " do_populate_license_readme; "

# Setup sysctl to release the closed/orphaned tcp connections immediately.
inherit sysctl_config

IMAGE_CLASSES += " sysctl_config"
ROOTFS_POSTPROCESS_COMMAND += " sysctl_config_set_tcp_fin_timeout; sysctl_config_set_core_pattern; sysctl_config_eth_options; "

python () {
    # Set GIT_COMMIT_COUNT and KAS_OS_VERSION fields
    '''
        NOTE: for 'git' to succeed, the meta-kas directory must be designated as
        "safe" for a non-owner of the repo. This is done using git config. Example:
            git config --global --add safe.directory /timesys/yocto-kas/layers/meta-kas
        There may be a fix in Yocto for this in the future. The issue is discussed here:
        https://lists.openembedded.org/g/openembedded-core/message/165184.
    '''
    if d.getVar('KAS_OS_VERSION') == '':
        try:
            GIT_SRC_DIR = os.path.dirname(d.getVar("COREBASE", True)) +'/meta-kas'
            stream = os.popen(f'git -C {GIT_SRC_DIR} rev-list HEAD | wc -l')
            GIT_REV_KAS=stream.read().strip().zfill(5) # 5-digit revision number
            d.setVar('GIT_COMMIT_COUNT', GIT_REV_KAS)
            KAS_OS_VERSION = '{}.{}.{}.{}'.format(
                d.getVar('KAS_OS_VERSION_MAJOR'),
                d.getVar('KAS_OS_VERSION_MINOR'),
                d.getVar('KAS_OS_VERSION_MICRO'),
                GIT_REV_KAS
                )
            d.setVar('KAS_OS_VERSION', KAS_OS_VERSION)
        except Exception as exc:
            bb.fatal(str(exc))
        finally:
            bb.debug(1, "KAS_OS_VERSION set to ", d.getVar('KAS_OS_VERSION'))
    
    d.setVar('MACHINE_NAME_UPPER',  '{}'.format(d.getVar('MACHINE_NAME').upper()))

    # Backup those files/folders used when restoring the system back to factory defaults.
    # We want this to run as late as possible, in particular after systemd_preset_all and
    # reproducible_final_image_task shell functions are executed; whereafter the entire
    # file system gets populated.
    d.appendVar('IMAGE_PREPROCESS_COMMAND', 'image_rootfs_create_default_dir')
}

# Define kas-data partition directory
KAS_DATA_PARTITION_DIR="${DEPLOY_DIR_IMAGE}/kas-data"

SRC_URI:append := " \
    file://sda_fstab \
    file://README \
    "

INHIBIT_SWUPDATE_ADD_SRC_URI="1"

# Tell wic not to update fstab -- we use this in kas-data partition
WIC_CREATE_EXTRA_ARGS="--no-fstab-update"

do_image_wic_prep() {
    install -d ${DEPLOY_DIR_IMAGE}
}
addtask do_image_wic_prep before do_image_wic

# fstab is overwritten by USB build before building wic, so back it up.
do_backup_fstab() {
    install "${IMAGE_ROOTFS}/etc/fstab" "${WORKDIR}/mmc_fstab"
}
ROOTFS_POSTPROCESS_COMMAND += " do_backup_fstab;"


# Create bare directory for kas-data partition before SWU files are generated
# The swu files themselves are inserted after the WIC image is generated
# and since the wic commands are limited, it is more convenient to place whatever
# files are available aready with the mechanism below.
do_image_data_partition_before_swu() {
    rm -rf "${KAS_DATA_PARTITION_DIR}"
    install -d "${KAS_DATA_PARTITION_DIR}/kas" "${KAS_DATA_PARTITION_DIR}/system/1"    \
       "${KAS_DATA_PARTITION_DIR}/system/2" "${KAS_DATA_PARTITION_DIR}/system/3"       \
       "${KAS_DATA_PARTITION_DIR}/system/mmc" "${KAS_DATA_PARTITION_DIR}/system/sda"

    # create following KAS runtime specific directories in the data partition under /kas-data/kas 
    install -d "${KAS_DATA_PARTITION_DIR}/kas/application" "${KAS_DATA_PARTITION_DIR}/kas/crashdump" \
        "${KAS_DATA_PARTITION_DIR}/kas/log" "${KAS_DATA_PARTITION_DIR}/kas/userdata"

    install -T -m 0644 "${WORKDIR}/sda_fstab" "${KAS_DATA_PARTITION_DIR}/system/sda/fstab"
    install -T -m 0644 "${WORKDIR}/mmc_fstab" "${KAS_DATA_PARTITION_DIR}/system/mmc/fstab"

    install -T -m 0644 "${DEPLOY_DIR_IMAGE}/${BOOTFILES_DIR_NAME}/cmdline-usb.txt"      \
           "${KAS_DATA_PARTITION_DIR}/system/sda/cmdline.txt"
    install -T -m 0644 "${DEPLOY_DIR_IMAGE}/${BOOTFILES_DIR_NAME}/cmdline-mmc.txt"      \
           "${KAS_DATA_PARTITION_DIR}/system/mmc/cmdline.txt"   
}

# Note: The implementation of the below methods are in kas-image.bb and kas-image-usb-common.inc files
addtask do_image_data_partition_before_swu before do_image_wic after do_write_wks_template

# We are storing the kernel image in the rootfs partition so remove it from the boot partition
IMAGE_BOOT_FILES:remove = "${KERNEL_IMAGETYPE}"

# Add a buffer to the image's rootfs size (Kbytes). 
# Partition size is 1.95G, use 1.5G for fs, add 0.45G as extra size for swupdate to work.
IMAGE_ROOTFS_SIZE="1536000"
IMAGE_ROOTFS_EXTRA_SIZE="460800"
