#!/bin/bash

# Init script for PCMM startup processes

KAS=/home/kas/kas
KAS_LOG_DIR=/kas-data/kas/log
SERVICE_LOG_FILE=/var/log/pcmm-init.log

echo "$(date): Starting KAS system intialization" | tee ${SERVICE_LOG_FILE} 2>&1

# Display '3'(Code: 0xB0) or 'J' (0xE1) on 7-seg depending on the boot source. 
# This code indicates the user that the system process manager started.
boot_partition_mnt_dir=$(findmnt -M /boot --noheadings --output SOURCE)
if [ "${boot_partition_mnt_dir}" == "/dev/sda1" ]; then
    # We're booting off from the Recovery USB flash
    echo "$(date): Booting off from Recovery USB flash" | tee -a ${SERVICE_LOG_FILE} 2>&1
    i2cset -y 2 0x74 0x03 0xE1
else
    # We're booting off from the eMMC internal flash.
    echo "$(date): Booting off from the main board eMMC flash" | tee -a ${SERVICE_LOG_FILE} 2>&1
    i2cset -y 2 0x74 0x03 0xB0
fi

if [ ! -d "${KAS}" ]; then
    echo "$(date): KAS runtime is not installed at ${KAS}." | tee -a ${SERVICE_LOG_FILE} 2>&1
    exit 0
fi

if [ ! -d "${KAS_LOG_DIR}" ]; then
    echo "$(date): KAS runtime log directory not present at ${KAS_LOG_DIR}" | tee -a ${SERVICE_LOG_FILE} 2>&1
    KAS_LOG_DIR=/var/log
fi

# Set permissions for the KAS runtime installation directory
chown -R kas.kas ${KAS}

# Start the KAS system initialization script. 
# This script launches the KAS runtime. Also, verifies the upgrade status and initiate the rollback when necessary.
echo "$(date): Starting ${KAS}/scripts/kas_sysinit.py script" | tee -a  ${SERVICE_LOG_FILE} 2>&1
python3 ${KAS}/scripts/kas_sysinit.py > ${KAS_LOG_DIR}/_kas_sysinit.log  2>&1
exit_code=$?
echo "$(date): KAS system intialization complete. exit_code=${exit_code}" | tee -a ${SERVICE_LOG_FILE} 2>&1

if [ $exit_code -ne 0 ]; then
    # fw_printenv returns an expression like "name=value".
    # Use export to set a variable to the evaluated expression returned by fw_printenv.
    export `fw_printenv broken_root_partition`

    if [ "${boot_partition_mnt_dir}" == "/dev/sda1" ]; then
        echo "$(date): Booting off recovery drive. Ignoring kas_sysinit.py exit code." | tee -a ${SERVICE_LOG_FILE} 2>&1
        exit_code=0
    elif [ "${broken_root_partition}" != "" ]; then
        echo "$(date): broken_root_partition is ${broken_root_partition}. Ignoring kas_sysinit.py exit code." | tee -a ${SERVICE_LOG_FILE} 2>&1
        exit_code=0
    fi
fi

RTC_I2CADDR=0x51
RTC_RAM_ADDR=0x3
RTC_RAM_BYTE_FILE=/sys/bus/i2c/devices/2-0051/pcf85063_nvram0/nvmem

if [ $exit_code -ne 0 ]; then
    # Save the current root partition to the broken_root_partition environment variable.
    root_partition_mnt_dir=$(findmnt -M / --noheadings --output SOURCE)
    fw_setenv broken_root_partition $root_partition_mnt_dir

    # max out the bootcount to force a reboot to the inactive partition
    printf '\xb5' > $RTC_RAM_BYTE_FILE
    reboot
else
    # clear the bootcount
    printf '\xb0' > $RTC_RAM_BYTE_FILE
fi

exit $exit_code
