import asyncio
import os
import subprocess
import sys

if sys.platform == 'win32':
	import nt

_DEFAULT_LIMIT = asyncio.streams._DEFAULT_LIMIT * 16

class RunOutput():
	def __init__(self, returncode, stdout, stderr):
		self.returncode = returncode
		self.stdout = stdout
		self.stderr = stderr

async def _read_stream(stream, callback):
	while True:
		try:
			line = await stream.readline()
		except ValueError as exc:
			callback(f'warning: exeption in asynchio_tee._read_stream:\n')
			callback(str(exc))
			return
			
		if line:
			callback(line)
		else:
			break
		
async def _stream_subprocess(cmd
	, stdin = None
	, echo = False
	, suppressOutput = False
	, stdoutOnly = False
	, useCaseSensitiveEnvVars = False
	, codec = 'utf-8'
	, shell = True
	, stdoutLabel = ''
	, stderrLabel = 'stderr: '
	, limit = _DEFAULT_LIMIT
) -> RunOutput:
	if sys.platform == 'win32':
		environ = os.environ
		if useCaseSensitiveEnvVars:
			# Pass in the unadulterated environment. The keys for os.environ are converted to upper case.
			environ = nt.environ
		
		platform_settings = {
			'env': environ,
		}
	else:
		platform_settings = {
			'executable': '/bin/bash',
		}
	
	# convert cmd to a str - copied from subprocess
	if not shell:
		# for not shell, only list-like objects are supported
		assert not isinstance(cmd, (str, bytes, os.PathLike))
	
	if isinstance(cmd, str):
		cmdLine = cmd
	elif isinstance(cmd, (bytes, os.PathLike)):
		cmdLine = subprocess.list2cmdline([cmd])
	else:
		cmdLine = subprocess.list2cmdline(cmd)

	if echo:
		print(cmdLine)
	
	if shell:
		p = await asyncio.create_subprocess_shell(
			cmdLine,
			stdin = stdin,
			stdout = asyncio.subprocess.PIPE,
			stderr = asyncio.subprocess.PIPE,
			limit = limit,
			**platform_settings
		)
	else:
		p = await asyncio.create_subprocess_exec(
			*cmd,
			stdin = stdin,
			stdout = asyncio.subprocess.PIPE,
			stderr = asyncio.subprocess.PIPE,
			limit = limit,
			**platform_settings
		)
	
	out = []
	if stdoutOnly:
		err = out
	else:
		err = []

	def tee(line, sink, pipe, codec
		, label = ''
	):
		try:
			line = line.decode(codec)
		except UnicodeDecodeError as exc:
			line = 'in asyncio_tee, failed to decode line: %s' % line
		
		line = line.rstrip()
		
		sink.append(line)
		if not suppressOutput:
			print(label, line
				, file = pipe
			)
	
	await asyncio.wait([
		_read_stream(p.stdout, lambda l: tee(l, out, sys.stdout, codec, label=stdoutLabel)),
		_read_stream(p.stderr, lambda l: tee(l, err, sys.stderr, codec, label=stderrLabel)),
	])

	return RunOutput(await p.wait(), out, err)


def run(cmd
	, stdin = None
	, echo = True
	, suppressOutput = False
	, stdoutOnly = False
	, useCaseSensitiveEnvVars = False
	, codec = 'utf-8'
	, shell = True
	, stdoutLabel = ''
	, stderrLabel = 'stderr: '
):
	loop = asyncio.get_event_loop()
	
	runOutput = _stream_subprocess(cmd
		, stdin = stdin
		, echo = echo
		, suppressOutput = suppressOutput
		, stdoutOnly = stdoutOnly
		, useCaseSensitiveEnvVars = useCaseSensitiveEnvVars
		, codec = codec
		, shell = shell
		, stdoutLabel = stdoutLabel
		, stderrLabel = stderrLabel
	)
	
	result = loop.run_until_complete(runOutput)
	
	exitCode = result.returncode
	out = '\n'.join(result.stdout)
	err = '\n'.join(result.stderr)
	
	return exitCode, out, err

def _test_run():
	cmd = [ "ping", "localhost" ]
	exitCode, out, err = run(cmd
		, suppressOutput = False
		, echo = True
	)
	
	print('exitCode: %d' % exitCode)
	if len(out):
		print('stdout:\n%s\n' % out)
	
	if len(err):
		print('stderr:\n%s\n' % err)
	
if __name__ == '__main__':
	_test_run()
	