r'''
Copyright 2023 by Kollmorgen Corporation.  All rights reserved.

This script allows to build the OS & Recovery USB images.
'''

import argparse
from collections import namedtuple
import json
import os
import pathlib
import re
import shutil
import sys
import time
import urllib.request

import asyncio_tee
import util


BB_MMC_IMAGE_NAME      = 'kas-image'
BB_RECOVERY_IMAGE_NAME = 'kas-image-usb'
MACHINE_NAME           = 'kas-pcmm2g'

KAS_JENKINS_URL              = 'http://sbaws-vm01.ptgroup.ptgnet.us:8080'
KAS_JENKINS_BUILD_JOB_NAME   = 'kas_build_artifacts_trunk'
KAS_BUILD_ARTIFACTS_BASE_URL = f'{KAS_JENKINS_URL}/job/{KAS_JENKINS_BUILD_JOB_NAME}'
KAS_LATEST_BUILD_URL         = f'{KAS_BUILD_ARTIFACTS_BASE_URL}/lastSuccessfulBuild/api/json'

PCMM2G_IMAGE_BASENAME                = 'KAS-PCMM2G-Cx'
PCMM2G_IMAGE_BASENAME_WITH_FLASHSIZE = f'{PCMM2G_IMAGE_BASENAME}-08'
PCMM2G_IMAGE_FILE_RT_BASENAME        = f'{PCMM2G_IMAGE_BASENAME_WITH_FLASHSIZE}-RT-'

bb_build_dir            = 'build-kas'
bb_user_config_filename = 'build_user.conf'
build_version_filepath  = None


def bb_build_dir_set(build_dir) -> None:
	global bb_build_dir
	bb_build_dir = build_dir


def bb_buid_dirpath_get() -> pathlib.Path:
	return util.root_dir / bb_build_dir


def bb_build_tmp_dir_get() -> pathlib.Path:
	return bb_buid_dirpath_get() / 'tmp'


def bb_build_deploy_dir_get() -> pathlib.Path:
	return bb_build_tmp_dir_get() / 'deploy'


def bb_build_pcmm2g_images_dir_get() -> pathlib.Path:
	return bb_build_deploy_dir_get() / 'images' / 'kas-pcmm2g'


def bb_user_config_filepath_get() -> pathlib.Path:
	return bb_build_tmp_dir_get() / bb_user_config_filename


def build_version_filepath_get() -> pathlib.Path:
	if build_version_filepath is None:
		return bb_buid_dirpath_get() / 'version.json'

	return build_version_filepath


def build_version_filepath_set(version_filepath_new: str | pathlib.Path) -> None:
	global build_version_filepath

	if isinstance(version_filepath_new, str):
		version_filepath_new = pathlib.Path(version_filepath_new)

	build_version_filepath = version_filepath_new

	version_filepath_new.parent.mkdir(
		parents = True,
		exist_ok = True,
	)


def bitbake_recipe_clean_build(
	recipe_name,
	verbose = False,
) -> int:
	verbose_arg = ' --verbose' if verbose else ''
	build_cmd   = f'source {util.root_dir}/setup-environment {bb_buid_dirpath_get()}'
	build_cmd  += f' && bitbake{verbose_arg} -c clean {recipe_name}'

	exit_code, stdout, stderr = execute_command(cmd = build_cmd)
	return exit_code


def build_os_only_image(
	os_version	= None,
	force_build = False,
	verbose		= False,
) -> str:
	print('Building OS image (.swu)')

	err_msg, config_filepath = generate_bitbake_user_config_file(
		install_runtime = False,
		os_version		= os_version,
	)
	if err_msg is not None:
		return err_msg

	exit_code = execute_bitbake_build(
		recipe_name      = BB_MMC_IMAGE_NAME,
		user_config_file = config_filepath,
		force_build      = force_build,
		verbose          = verbose,
	)

	if exit_code != 0:
		err_msg = f'Failed to build OS only image. exit_code={exit_code}'
		return err_msg


def build_os_sdk(verbose = False) -> str | None:
	print('Building OS image (.swu)')

	err_msg, config_filepath = generate_bitbake_user_config_file(install_runtime = False)
	if err_msg is not None:
		return err_msg

	exit_code = execute_bitbake_build(
		recipe_name      = 'kas-image',
		user_config_file = config_filepath,
		verbose          = verbose,
		command          = 'populate_sdk',
	)

	if exit_code != 0:
		err_msg = f'Failed to build the SDK. exit_code={exit_code}'
		return err_msg


def build_recovery_os_with_runtime_image(
	branch_name = None,
	os_version	= None,
	force_build = False,
	verbose		= False,
) -> str:
	print('Building Recovery USB image (.wic.bz2)')

	details = get_kas_runtime_latest_build_details(branch_name = branch_name, verbose = verbose)
	if details.err_msg is not None:
		return details.err_msg

	err_msg, config_filepath = generate_bitbake_user_config_file(
		install_runtime            = True,
		os_version				   = os_version,
		kas_jenkins_build_number   = details.build_id,
		kas_version                = details.kas_version,
		kas_install_file_checksum  = details.pcmm2g_file_crc,
		kas_jenkins_build_job_name = KAS_JENKINS_BUILD_JOB_NAME,
	)

	if err_msg is not None:
		return err_msg

	exit_code = execute_bitbake_build(
		recipe_name      = BB_RECOVERY_IMAGE_NAME,
		user_config_file = config_filepath,
		force_build      = force_build,
		verbose          = verbose,
	)

	if exit_code != 0:
		err_msg = f'Failed to build recovery image. exit_code={exit_code}'
		return err_msg


def build_version_file_write(
	version_file_path,
	branch_name = None,
	os_version	= None,
	verbose		= False,
) -> str:
	version_file_path = pathlib.Path(version_file_path)
	parent            = version_file_path.parent

	if verbose:
		print(f'Version file path: {version_file_path}')

	if not parent.is_dir():
		err_msg = f'Config file\'s parent directory not a valid directory : {parent}'
		return err_msg

	if os_version is None:
		os_version = util.os_version_get()

	details = get_kas_runtime_latest_build_details(branch_name = branch_name, verbose = verbose)

	if details.err_msg is not None:
		return details.err_msg

	util.version_file_write(version_file_path, os_version, details.kas_version)


def clean_build_dirs(verbose = False) -> str:
	bb_buid_dirpath = bb_buid_dirpath_get()
	print(f'Cleaning build area. build_dir: {bb_buid_dirpath}')
	clean_dirs = [
		bb_buid_dirpath / 'sstate-cache',
		bb_buid_dirpath / 'tmp',
	]

	for clean_dir in clean_dirs:
		clean_dir = clean_dir.resolve()
		if not clean_dir.exists():
			continue

		if not clean_dir.is_dir():
			err_msg = f'clean_build_dirs(): {clean_dir} is not a directory'
			return err_msg

		if verbose:
			print(f'Removing {clean_dir}')

		shutil.rmtree(clean_dir)


def execute_bitbake_build(
	recipe_name,
	user_config_file,
	verbose     = False,
	command     = None,
	force_build = False,
) -> int:
	build_time_start = time.time()
	if force_build and command is None:
		exit_code = bitbake_recipe_clean_build(recipe_name, verbose)
		if exit_code != 0:
			print(f'Failed to execute the clean build. exit_code: {exit_code}')
			return exit_code

	verbose_arg = ' --verbose' if verbose else ''
	command_arg = '' if command is None else f' -c {command}'
	bitbake_cmd = f'bitbake --read={user_config_file}{verbose_arg} {recipe_name}{command_arg}'
	build_cmd   = f'source {util.root_dir}/setup-environment {bb_buid_dirpath_get()}'
	build_cmd  += f' && {bitbake_cmd}'

	exit_code, stdout, stderr = execute_command(build_cmd)
	'''
	Note: Bitbake builds performed right after cleaning the image is returning
	a non-zero exit code even when the build it successful. The subsequent build
	usually succeeds (provided there are no real errors). 	
		
	Bitbake is printing following warning/error messages in this case. 
		ERROR: When reparsing /mnt/yocto/km-github/yocto-kas-pcmm2g/build-kas/../layers/meta-kas/recipes-core/images/kas-image.bb:do_image_ext4,
			the basehash value changed from 9222cbd21faedf7ca281ff4b9758268f6b8568a0edabfd0a7285f1583cfd5442 to
			e74d8095198eae4e572213857d53b6670edacf8a879b83b1f9bdd50cb4e671e1. The metadata is not deterministic and this needs to be fixed.
		ERROR: The following commands may help:
		ERROR: $ bitbake kas-image -cdo_image_ext4 -Snone
		ERROR: Then:
		ERROR: $ bitbake kas-image -cdo_image_ext4 -Sprintdiff

	We do not understand the reason for these error message.
	Hence, addng a temporary temporar to force another build if the first build failed.
	'''
	if exit_code != 0 and force_build:
		print(f'**** Rettempting to build {recipe_name} image  ****')
		exit_code, stdout, stderr = execute_command(build_cmd)

	build_time_elapsed = time.time() - build_time_start
	print(f'Total elapsed time for building {recipe_name} recipe: {build_time_elapsed:.0f} secs. ')

	return exit_code


def execute_command(cmd: str) -> int:
	print(f'Executing the command: "{cmd}"')
	cmd_start_time            = time.time()
	exit_code, stdout, stderr = asyncio_tee.run(
		cmd            = cmd,
		echo           = False,
		suppressOutput = False,
	)
	elapsed_time = time.time() - cmd_start_time
	print(f'Elapsed time for command execution: {elapsed_time:.0f} secs. exit_code: {exit_code}')

	return exit_code, stdout, stderr


KAS_Build_Details = namedtuple(
	"KAS_Build_Details",
	"err_msg build_id kas_version, pcmm2g_file_crc",
	defaults = [None, None, None, None]
)


def get_kas_runtime_latest_build_details(
	branch_name: str = None,
	verbose: bool	 = False,
) -> KAS_Build_Details:
	
	global KAS_JENKINS_BUILD_JOB_NAME
	global KAS_BUILD_ARTIFACTS_BASE_URL
	global KAS_LATEST_BUILD_URL

	KAS_JENKINS_BUILD_JOB_NAME = \
		'kas_build_artifacts_trunk' if branch_name is None else \
		fr'kas_build_artifacts/job/Releases%252F{branch_name}%252Fdev%252Fbuild%252FJenkins%252Fkas_build_artifacts'

	KAS_BUILD_ARTIFACTS_BASE_URL = f'{KAS_JENKINS_URL}/job/{KAS_JENKINS_BUILD_JOB_NAME}'
	KAS_LATEST_BUILD_URL         = f'{KAS_BUILD_ARTIFACTS_BASE_URL}/lastSuccessfulBuild/api/json'

	try:
		url_response_str = None
		with urllib.request.urlopen(KAS_LATEST_BUILD_URL) as url_response:
			url_response_str = url_response.read()
	except urllib.error.HTTPError as exc:
		err_msg = f'failed to open URL {KAS_LATEST_BUILD_URL} : {str(exc)}'
		return KAS_Build_Details(err_msg)

	build_info_json = json.loads(url_response_str)

	json_formatted_str = json.dumps(build_info_json, indent = 2)

	debug = False
	if debug:
		print(f'url_response_json=\n{json_formatted_str}')

	build_id = build_info_json.get("id")
	if build_id is None:
		err_msg = 'Failed to get build id from the URL response'
		return KAS_Build_Details(err_msg)

	artifacts = build_info_json.get('artifacts')
	if artifacts is None:
		err_msg = 'Latest build response doesn\'t contain artifacts'
		return KAS_Build_Details(err_msg)

	pcmm2g_rt_filename_pattern = re.compile(f'^{PCMM2G_IMAGE_FILE_RT_BASENAME}(\d+\.\d+\.\d+\.\d+)\.zip$')

	kas_version               = None
	kas_artifact_relative_url = None

	for artifact_info in artifacts:
		if verbose:
			print(f'info: {artifact_info}')
		artifact_filename = artifact_info.get('fileName')
		if artifact_filename is None:
			continue
		result = pcmm2g_rt_filename_pattern.match(artifact_filename)
		if not result:
			continue

		search_groups = result.groups()
		if len(search_groups) != 1:
			break

		kas_version               = search_groups[0]
		kas_artifact_relative_url = artifact_info.get('relativePath')
		break

	if kas_version is None or kas_artifact_relative_url is None:
		err_msg = 'Failed to get latest KAS runtime version details'
		return KAS_Build_Details(err_msg)

	kas_build_artifacts_latest_url = f'{KAS_BUILD_ARTIFACTS_BASE_URL}/{build_id}/artifact/{kas_artifact_relative_url}'
	kas_artifact_sha256_cmd        = f'curl -s {kas_build_artifacts_latest_url} | sha256sum'

	exit_code, response, stderr = asyncio_tee.run(
		cmd            =kas_artifact_sha256_cmd,
		echo           =verbose,
		suppressOutput =not debug)
	if exit_code != 0:
		err_msg = f'Failed to get sha256sum using curl command'
		return KAS_Build_Details(err_msg)

	search_str      = ' -'
	pcmm2g_file_crc = response.partition(search_str)[0].strip()

	if verbose:
		print(f'KAS latest runtime details:')
		print(
			f'jenkins_build_id: {build_id}, kas_version: {kas_version},  sha256sum: {pcmm2g_file_crc}'
		)

	return KAS_Build_Details(None, build_id, kas_version, pcmm2g_file_crc)


def generate_bitbake_user_config_file(
	install_runtime: bool		    = False,
	os_version: str				    = None,	
	kas_version: str                = None,
	kas_jenkins_build_job_name: str = None,
	kas_jenkins_build_number: str   = None,
	kas_install_file_checksum: str  = None,
) -> tuple[str, pathlib.Path]:

	if kas_jenkins_build_job_name is None:
		kas_jenkins_build_job_name = KAS_JENKINS_BUILD_JOB_NAME
		
	bb_user_config_filepath = bb_user_config_filepath_get()

	# Delete configuration file, if it's already present.
	bb_user_config_filepath.unlink(missing_ok = True)

	os.makedirs(bb_user_config_filepath.parent, exist_ok = True)

	print(
		f'Creating configuration file: {bb_user_config_filepath}. install_runtime = {install_runtime}, os_version={os_version}'
	)
	if install_runtime:
		print(
			f'kas_version={kas_version}, build_id: {kas_jenkins_build_number}, crc: {kas_install_file_checksum}'
		)

	install_runtime_val = "1" if install_runtime else "0"

	try:
		# write configuration file entries
		with open(bb_user_config_filepath, 'w') as config_file:
			config_file.write(f'INSTALL_KAS_RUNTIME="{install_runtime_val}"\n')

			if os_version is not None:
				config_file.write(f'KAS_OS_VERSION="{os_version}"\n')
			
			if install_runtime:
				config_file.write(f'KAS_JENKINS_BUILD_JOB_NAME="{kas_jenkins_build_job_name}"\n')
				config_file.write(f'KAS_JENKINS_BUILD_NUMBER="{kas_jenkins_build_number}"\n')				
				config_file.write(f'KAS_RUNTIME_VERSION="{kas_version}"\n')
				config_file.write(f'KAS_SRC_URI_CHECKSUM="{kas_install_file_checksum}"\n')

			return None, bb_user_config_filepath
	except Exception as exc:
		err_msg = f'Exception occurred when updating bitbake configuration file- {exc}'
		return err_msg, None

	return 'failed to generate user build config file', None


def main(argv):

	parser = argparse.ArgumentParser(description = __doc__, )

	parser.add_argument(
		'--build-dir',
		dest     = 'build_dir',
		default  = None,
		required = False,
		help     = 'Set the build directory.',
	)

	parser.add_argument(
		'--build-os-image',
		dest     = 'build_os_image',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'Builds OS with no KAS runtime included in the image (.swu)',
	)

	parser.add_argument(
		'--build-recovery-image',
		dest     = 'build_recov_image',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'Builds Recovery USB image with KAS runtime (.wic.bz2)',
	)

	parser.add_argument(
		'--build_sdk',
		dest     = 'build_sdk',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'Builds the SDK.',
	)

	parser.add_argument(
		'--branch-name',
		dest     = 'branch_name',
		default  = None,
		required = False,
		help     = 'Set the branch name (Ex. 04.01).',
	)

	parser.add_argument(
		'--clean',
		dest     = 'clean',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'Clear build directories',
	)

	parser.add_argument(
		'-f',
		'--force-build',
		dest     = 'force_build',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'Force generating the image',
	)

	parser.add_argument(
		'--os-version',
		dest     = 'os_version',
		default  = None,
		required = False,
		help     = 'Use this OS version instead of recalculating it.',
	)

	parser.add_argument(
		'-w',
		'--write-version-file',
		dest     = 'version_file_write',
		nargs    = '?',
		default  = None,
		const    = '',
		required = False,
		help     = '''
		   Generates a configuration file with OS & KAS runtime version details included.

		   Note: This option requires access to the Kollmorgen's KAS Jenkin Server and 
		   Kollmorgen GitHub repository.
		   ''',
	)

	parser.add_argument(
		'-v',
		'--verbose',
		dest     = 'verbose',
		default  = False,
		action   = 'store_true',
		required = False,
		help     = 'enable the verbose ouptut for the bitbake command',
	)

	args               = parser.parse_args()
	verbose            = args.verbose
	build_os_image     = args.build_os_image
	build_recov_image  = args.build_recov_image
	clean_build        = args.clean
	force_build        = args.force_build
	build_dir          = args.build_dir
	build_sdk          = args.build_sdk
	version_file_write = args.version_file_write
	os_version         = args.os_version

	if build_dir is not None:
		bb_build_dir_set(build_dir)

	if clean_build:
		err_msg = clean_build_dirs(verbose = verbose)
		if err_msg is not None:
			return err_msg

	if version_file_write is not None:
		if len(version_file_write):
			build_version_filepath_set(version_file_write)

		version_filepath = build_version_filepath_get()
		err_msg = build_version_file_write(
			version_filepath, 
			branch_name	= args.branch_name,
			verbose		= verbose,
			os_version	= os_version,
		)
		if err_msg is not None:
			return err_msg

	if build_os_image:
		err_msg = build_os_only_image(
			os_version	= os_version,
			force_build = force_build,
			verbose     = verbose,
		)
		if err_msg is not None:
			return err_msg

	if build_recov_image:
		err_msg = build_recovery_os_with_runtime_image(
			branch_name = args.branch_name,
			os_version	= os_version,
			force_build = force_build,
			verbose     = verbose,
		)
		if err_msg is not None:
			return err_msg

	if build_sdk:
		err_msg = build_os_sdk(verbose = verbose)
		if err_msg is not None:
			return err_msg

if __name__ == '__main__':
	err_msg = main(sys.argv)
	sys.exit(err_msg)
