'''Generate Yocto build artifacts.
'''
import argparse
import collections
import json
import os
import pathlib
import shutil
import sys

import util

ValueOrError = collections.namedtuple(
	'ValueOrError',
	field_names = ('value', 'error'),
	defaults = (None, None)
)

kas_artifact_file_base = 'KAS-PCMM2G-Cx-08'

kas_tgz_file_name_format = kas_artifact_file_base + '-RT-{kas_version}.tgz'

license_sources_zip_name_format = 'KAS-PCMM2G-OS-License-Sources-{os_version}.zip'

boot_image_symbolic_link_file_name = kas_artifact_file_base + '-bootimage.swu'
boot_image_zip_name_format = kas_artifact_file_base + '-bootimage-{boot_image_version}.zip'

cve_report_file_name_format = kas_artifact_file_base + '-OS-{os_version}.json'

os_image_zip_name_format = kas_artifact_file_base + '-OS-{os_version}.zip'
os_image_plus_kas_zip_name_format = kas_artifact_file_base + '-OS-{os_version}-RT-{kas_version}.zip'
os_swu_file_name_format = kas_artifact_file_base + '-OS-{os_version}.swu'

recovery_file_name_format = kas_artifact_file_base + '-Recovery-OS-{os_version}-RT-{kas_version}.wic.bz2'

yocto_sources_zip_file_format = 'KAS-PCMM2G-OS-Build-Sources-{os_version}.zip'

build_dir_default = 'build-kas'

def artifacts_create(
	artifact_dir,
	export_dir,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	os.makedirs(artifact_dir, exist_ok = True)
	if export_dir is not None:
		os.makedirs(export_dir, exist_ok = True)

	# Create artifacts.
	error = None

	for artifact_name, artifact_data in artifact_map.items():
		function = artifact_data['function']
		file_name_format = artifact_data['file_name_format']
		file_name = file_name_format.format(**version_dict)

		error = function(
			artifact_dir,
			file_name,
			version_dict,
			use_existing_files = use_existing_files,
			build_dir = build_dir,
		)
			
		if error is not None:
			return error
	

def artifacts_create_and_export(
	artifact_dir,
	version_dict,
	export_dir = None,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	# Add the boot_image_version to version_dict. We get this from
	# the link that points to the boot image swu file.
	value_or_error = boot_image_version_get(
		build_dir = build_dir,
	)

	if value_or_error.error is not None:
		return value_or_error.error
	
	version_dict['boot_image_version'] = value_or_error.value
	
	build_dir_full = util.root_dir / build_dir
	if not build_dir_full.is_dir():
		return f'not a directory: {build_dir_full}'

	error = artifacts_create(
		artifact_dir,
		export_dir,
		version_dict,
		use_existing_files = use_existing_files,
		build_dir = build_dir,
	)

	if error is not None:
		return error

	if export_dir is not None:
		artifacts_export(
			artifact_dir, 
			export_dir, 
			version_dict,
			use_existing_files = use_existing_files
		)
	
	return None

def artifacts_export(
	artifact_dir,
	export_dir,
	version_dict,
	use_existing_files = False,
):
	if export_dir is None:
		return

	artifact_dir = pathlib.Path(artifact_dir)
	export_dir = pathlib.Path(export_dir)

	os.makedirs(export_dir, exist_ok = True)

	for artifact_name, artifact_data in artifact_map.items():
		file_name_format = artifact_data['file_name_format']
		file_name = file_name_format.format(**version_dict)

		export_file_path = export_dir / file_name
		if export_file_path.is_file():
			if use_existing_files:
				print(f'file already exists in {export_dir}: {file_name}')
				continue

		print(f'copying {file_name} from {artifact_dir} to {export_dir}')
		src = artifact_dir / file_name
		shutil.copy(src, export_dir)


def boot_image_only_zip(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	value_or_error = boot_image_swu_file_path_get(build_dir)

	if value_or_error.error is not None:
		return value_or_error.error

	swu_file_path = value_or_error.value
	swu_file_name = swu_file_path.name

	return swu_file_only_zip(
		artifact_dir,
		artifact_file_name,
		swu_file_name,
		'boot image',
		use_existing_files = use_existing_files,
		build_dir = build_dir,
	)

def boot_image_swu_file_path_get(
	build_dir = build_dir_default,
):
	build_tmp_dir = util.root_dir / build_dir / 'tmp'
	image_deploy_dir = build_tmp_dir / 'deploy' / 'images' / 'kas-pcmm2g'

	link_file_path = image_deploy_dir / boot_image_symbolic_link_file_name

	if not link_file_path.exists():
		return ValueOrError(
			error = f'.swu file does not exist: {link_file_path}',
		)
	
	if not link_file_path.is_symlink():
		return ValueOrError(
			error = f'.swu file is not a symbolic link: {link_file_path}',
		)
	
	return ValueOrError(
		value = link_file_path.readlink(),
	)

def boot_image_version_get(
	build_dir = build_dir_default,
):
	value_or_error = boot_image_swu_file_path_get(build_dir)

	if value_or_error.error is not None:
		return value_or_error

	swu_file_path = value_or_error.value

	base_name = swu_file_path.with_suffix('').name
	name_parts = base_name.split('-')

	if len(name_parts) == 0:
		return ValueOrError(
			error = f'invalid file name: {swu_file_path}'
		)

	return ValueOrError(
		value = name_parts[-1],
	)


def license_sources_zip(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir = pathlib.Path(artifact_dir)
	zip_file_path = artifact_dir / artifact_file_name

	if zip_file_path.is_file():
		if use_existing_files:
			print(f'license source file already exists: {zip_file_path}')
			return

	# Zip up this entire directory:
	# {build_dir}/tmp/deploy/sources/aarch64-poky-linux
	source_dir = util.root_dir / build_dir / 'tmp' / 'deploy' / 'sources' / 'aarch64-poky-linux'

	source_items = {}

	for item in source_dir.iterdir():
		source_items[item.name] = {}

	util.zip_file_append(
		zip_file_path,
		source_items,
		open_mode = 'w',
		root_dir = source_dir,
	)

	print(f'generated license source file: {zip_file_path}')

def main(argv):
	parser = argparse.ArgumentParser(description = __doc__)

	parser.add_argument(
		'artifact_dir',
		help = 'path to the local artifact directory'
	)

	parser.add_argument(
		'--export_dir',
		dest = 'export_dir',
		default = None,
		required = False,
		help = 'path to the remote artifact directory'
	)

	parser.add_argument(
		'version_file_path',
		help = 'path to the version JSON file'
	)

	parser.add_argument(
		'-u',
		'--use_existing_files',
		dest = 'use_existing_files',
		default = False,
		action = 'store_true',
		required = False,
		help = 'use existing artifact files rather than overwrite them',
	)
	
	parser.add_argument(
		'--build_dir',
		dest = 'build_dir',
		default = build_dir_default,
		required = False,
		help = 'Set the build directory.',
	)

	args = parser.parse_args(args = argv[1:])

	version_dict = util.version_file_read(args.version_file_path)

	error = artifacts_create_and_export(
		args.artifact_dir,
		version_dict,
		export_dir = args.export_dir,
		use_existing_files = args.use_existing_files,
		build_dir = args.build_dir,
	)

	if error is None:
		return 0
	
	print(error)
	return 1

def os_image_only_zip(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	swu_file_name = os_swu_file_name_format.format(**version_dict)

	return swu_file_only_zip(
		artifact_dir,
		artifact_file_name,
		swu_file_name,
		'OS',
		use_existing_files = use_existing_files,
		build_dir = build_dir,
	)

def os_image_plus_kas_zip(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir = pathlib.Path(artifact_dir)

	swu_file_name = os_swu_file_name_format.format(**version_dict)
	swu_file_path = util.root_dir / build_dir / 'tmp' / 'deploy' / 'images' / 'kas-pcmm2g' / swu_file_name

	tgz_file_name = kas_tgz_file_name_format.format(**version_dict)

	missing_files = []

	tgz_file_path = util.root_dir / build_dir / 'tmp' / 'deploy' / 'tar' / 'kas-pcmm2g-arm_64bit_gcc_11_2_0-release.tgz'

	if not tgz_file_path.is_file():
		missing_files.append(str(tgz_file_path))

	if not swu_file_path.is_file():
		missing_files.append(str(swu_file_path))
	
	if len(missing_files):
		missing_files = '\n'.join(missing_files)
		return f'missing file(s):\n{missing_files}'

	zip_file_path = artifact_dir / artifact_file_name

	if zip_file_path.is_file():
		if use_existing_files:
			print(f'OS+RT firmware file already exists: {zip_file_path}')
			return
		
	source_items = {
		tgz_file_path: {
			'name': tgz_file_name,
		},
		swu_file_path: {
			'name': swu_file_path.name,
		},
	}

	util.zip_file_append(
		zip_file_path,
		source_items,
		open_mode = 'w',
	)

	print(f'generated OS+RT firmware file: {zip_file_path}')

def recovery_file_copy(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir = pathlib.Path(artifact_dir)
	bz2_file_name = artifact_file_name
	artifact_file_path = artifact_dir / bz2_file_name

	if artifact_file_path.is_file():
		if use_existing_files:
			print(f'Recovery firmware file already exists: {artifact_file_path}')
			return
		

	deploy_image_dir = util.root_dir / build_dir / 'tmp' / 'deploy' / 'images' / 'kas-pcmm2g'
	bz2_file_path = deploy_image_dir / bz2_file_name

	if not bz2_file_path.is_file():
		return f'file does not exist: {bz2_file_path}'
	
	shutil.copy(bz2_file_path, artifact_dir)
	print(f'copied Recovery firmware file: {artifact_file_path}')

def swu_file_only_zip(
	artifact_dir,
	artifact_file_name,
	swu_file_name,
	description,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir = pathlib.Path(artifact_dir)

	build_tmp_dir = util.root_dir / build_dir / 'tmp'
	image_deploy_dir = build_tmp_dir / 'deploy' / 'images' / 'kas-pcmm2g'

	swu_file_path = image_deploy_dir / swu_file_name

	missing_files = []

	if not swu_file_path.is_file():
		missing_files.append(str(swu_file_path))
	
	if len(missing_files):
		missing_files = '\n'.join(missing_files)
		return f'missing file(s):\n{missing_files}'

	zip_file_path = artifact_dir / artifact_file_name

	if zip_file_path.is_file():
		if use_existing_files:
			print(f'{description} firmware file already exists: {zip_file_path}')
			return
	
	# Create the config file that is required for OS-only firmware files.
	firmware_config_file_data = {
		"min_runtime_version": None,
	}

	firmware_config_file_path = image_deploy_dir / 'firmware_config_file.json'

	with open(firmware_config_file_path, 'w') as json_file:
		json.dump(
			firmware_config_file_data,
			json_file,
			indent = '\t',
			sort_keys = True,
		)

	source_items = {
		firmware_config_file_path: {
			'name': 'config.json',
		},
		swu_file_path: {
			'name': swu_file_path.name,
		},
	}

	util.zip_file_append(
		zip_file_path,
		source_items,
		open_mode = 'w',
	)

	print(f'generated {description} firmware file: {zip_file_path}')


def yocto_sources_zip(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir = pathlib.Path(artifact_dir)
	zip_file_path = artifact_dir / artifact_file_name

	if zip_file_path.is_file():
		if use_existing_files:
			print(f'Yocto build source file already exists: {zip_file_path}')
			return
		
	source_items = {
		'layers'					: {},
		'setup-environment'			: {},
		'layers/meta-kas/README.md'	: {
			'name': 'README.md'
		},
	}

	excluded = [
		'.git',
	]

	util.zip_file_append(
		zip_file_path,
		source_items,
		open_mode = 'w',
		root_dir = util.root_dir,
		excluded = excluded,
	)

	print(f'generated Yocto build source file: {zip_file_path}')

def cve_report_copy(
	artifact_dir,
	artifact_file_name,
	version_dict,
	use_existing_files = False,
	build_dir = build_dir_default,
):
	artifact_dir 			= pathlib.Path(artifact_dir)
	cve_file_name			= artifact_file_name
	artifact_file_path_dest	= artifact_dir / cve_file_name

	deploy_image_dir = util.root_dir / build_dir / 'tmp' / 'deploy' / 'images' / 'kas-pcmm2g'
	cve_file_path_src 	 = deploy_image_dir / cve_file_name

	if not cve_file_path_src.is_file():
		return f'file does not exist: {cve_file_name}'
	
	shutil.copy(cve_file_path_src, artifact_dir)
	print(f'copied CVE report file: {artifact_file_path_dest}')


artifact_map = {
	'boot_image_only_zip': {
		'function'			: boot_image_only_zip,
		'file_name_format'	: boot_image_zip_name_format,
	},
	'os_image_only_zip': {
		'function'			: os_image_only_zip,
		'file_name_format'	: os_image_zip_name_format,
	},
	'os_image_plus_kas_zip': {
		'function'			: os_image_plus_kas_zip,
		'file_name_format'	: os_image_plus_kas_zip_name_format,
	},
	'yocto_sources_zip': {
		'function'			: yocto_sources_zip,
		'file_name_format'	: yocto_sources_zip_file_format,
	},
	'license_sources_zip': {
		'function'			: license_sources_zip,
		'file_name_format'	: license_sources_zip_name_format,
	},
	'recovery_file_copy': {
		'function'			: recovery_file_copy,
		'file_name_format'	: recovery_file_name_format,
	},
	'cve_report_copy': {
		'function'			: cve_report_copy,
		'file_name_format'	: cve_report_file_name_format,
	},
}

if __name__ == '__main__':
	exit_code = main(sys.argv)
	sys.exit(exit_code)
