import json
import os
import pathlib
import zipfile

op = os.path
opj = op.join

scripts_dir = pathlib.Path(__file__)
root_dir = scripts_dir.parents[3]

json_file_encoding = 'ascii'


class AutoUndoChdir():
	def __init__(
		self,
		new_dir: str,
		verbose: bool = False,
	):
		self.cur_dir = os.getcwd()
		self.new_dir = new_dir
		self.verbose = verbose

	def __enter__(self):
		if self.new_dir is not None:
			os.chdir(self.new_dir)

	def __exit__(self, exception_type, exception_value, exception_traceback):
		if self.new_dir is not None:
			os.chdir(self.cur_dir)


class PathEnvChanger:
	def __init__(self, new_path):
		self.old_path = os.environ.get('PATH', '')
		self.new_path = new_path
	
	def __enter__(self):
		environ = {
			'PATH': self.new_path,
		}
		
		os.environ.update(environ)

	def __exit__(self, exception_type, exception_value, exception_traceback):
		environ = {
			'PATH': self.old_path,
		}
		
		os.environ.update(environ)


def os_version_get():
	KAS_OS_VERSION_MAJOR = "1"
	KAS_OS_VERSION_MINOR = "0"
	KAS_OS_VERSION_MICRO = "0"

	GIT_SRC_DIR = root_dir / 'layers' / 'meta-kas'
	git_word_count_command = f'git -C {GIT_SRC_DIR} rev-list HEAD | wc -l'

	with os.popen(git_word_count_command) as stream:
		output = stream.read()

	# Use the output to form a 5-digit revision number.
	GIT_REV_KAS = output.strip().zfill(5)

	KAS_OS_VERSION = '{}.{}.{}.{}'.format(
		KAS_OS_VERSION_MAJOR,
		KAS_OS_VERSION_MINOR,
		KAS_OS_VERSION_MICRO,
		GIT_REV_KAS,
	)

	return KAS_OS_VERSION


def version_file_read(config_file_path):
	if config_file_path is None:
		config_dict = {
			'os_version': os_version_get(),
			'kas_version': '0.0.0.0'
		}

		return config_dict
	
	with open(
		config_file_path,
		'r',
		encoding = json_file_encoding,
	) as f:
		config_dict = json.load(f)
	
	return config_dict


def version_file_write(
	config_file_path,
	os_version = None,
	kas_version = '0.0.0.0',
):
	if os_version is None:
		os_version = os_version_get()

	config_dict = {
		'os_version': os_version,
		'kas_version': kas_version,
	}

	with open(
		config_file_path,
		'w',
		encoding = json_file_encoding,
	) as json_file:
		json.dump(
			config_dict,
			json_file,
			sort_keys = True,
			indent    = '\t',
		)
	
	print(f'generated configuration file: {config_file_path}')


def zip_file_append(
	zip_file_path,
	source_items,
	open_mode = 'x',
	root_dir = None,
	excluded = None,
):
	if excluded is None:
		excluded = []

	with zipfile.ZipFile(
		zip_file_path,
		open_mode,
		zipfile.ZIP_DEFLATED,
	) as archiveFile, AutoUndoChdir(
		root_dir
	) as auto_undo_chdir:
		for itemPath, itemData in source_items.items():
			itemPath = itemData.get('path', itemPath)
			itemName = itemData.get('name', itemPath)

			if op.isdir(itemPath):
				stack = [itemPath]
				while len(stack):
					curDir   = stack.pop(0)
					itemList = os.listdir(curDir)
					for item in itemList:
						if item in excluded:
							continue

						itemFull = opj(curDir, item)
						if op.isfile(itemFull):
							archiveFile.write(itemFull)
						elif op.isdir(itemFull):
							stack.append(itemFull)
							contentCount = len(os.listdir(itemFull))
							if contentCount == 0:
								# deal with empty directories - remove preceding / and add a / at the end
								path = op.relpath(itemFull, '/') + '/'

								zipInfo = zipfile.ZipInfo(path)
								archiveFile.writestr(zipInfo, '')
			else:
				archiveFile.write(itemPath, arcname = itemName)

